import os

from conversion import convert_bool

API_URL = os.environ.get('API_URL', 'https://zymbit.com/api/v1')
AUTH_ROOT = '/etc/zymbit/auth'
AUTH_TOKEN = os.environ.get('AUTH_TOKEN')
BOOTSTRAP_KEY = os.environ.get('BOOTSTRAP_KEY')
CHECK_HOSTNAME = convert_bool(os.environ.get('CHECK_HOSTNAME', 'true'))
CLIENT_ID_VERSION = os.environ.get('CLIENT_ID_VERSION')
PUBSUB_PING_INTERVAL = int(os.environ.get('PUBSUB_PING_INTERVAL', 300))
REGISTER_ENDPOINT = '/zymbots/register?key={BOOTSTRAP_KEY}'
WEBSOCKET_ENDPOINT = '/websocket_url'
WEBSOCKET_SEND_CLIENT_INFO = convert_bool(os.environ.get('WEBSOCKET_SEND_CLIENT_INFO', 'true'))
ZYMBIT_RUN_PATH = '/run/zymbit'
ZYMBIT_HOST_INFO_PATH = os.path.join(ZYMBIT_RUN_PATH, 'host_info')

CONNECT_PORT_9628_TCP_ADDR = os.environ.get('CONNECT_PORT_9628_TCP_ADDR', '0.0.0.0')
CONNECT_PORT_9628_TCP_PORT = int(os.environ.get('CONNECT_PORT_9628_TCP_PORT', 9628))

CONSOLE_MESSENGER_HOST = os.environ.get('CONSOLE_MESSENGER_HOST', CONNECT_PORT_9628_TCP_ADDR)
CONSOLE_MESSENGER_PORT = os.environ.get('CONSOLE_MESSENGER_PORT', CONNECT_PORT_9628_TCP_PORT)



import json
import uuid

from zymbit.util.time import timestamp

# different ways a data stream buffer is parsed in order to ship up
# NOTE: first one to send back an envelope wins, so order matters!
ENVELOPE_PARSERS = []


def get_parsed_envelope(params):
    action = 'data'

    if isinstance(params, dict):
        _action = params.pop('action', None)
        if _action is not None:
            action = _action

    if action == 'data' and 'key' not in params:
        params['key'] = 'sensor'

    if 'params' in params:
        params['action'] = action

        return json.dumps(params)

    return get_envelope(action, params)


def parse_json_envelope(buf):
    try:
        params = json.loads(buf)
    except ValueError:
        return None
    else:
        if isinstance(params, int):
            params = {
                'value': params,
            }

        return params
ENVELOPE_PARSERS.append(parse_json_envelope)


def parse_comma_equals(buf):
    """
    Parse a string of comma-delimited strings, that are each equal-delimited key/value pairs
    :param buf: string - buffer to be parsed
    :return: None - when no equal sign is found, JSON string envelop - when data is parsed
    """
    if '=' not in buf:
        return None

    parsed = {}
    unparsed = []

    # split at commas
    for token in buf.split(','):
        # get rid of outer spaces
        token = token.strip()

        if '=' not in token:
            unparsed.append(token)
            continue

        key, value = token.split('=')
        key = key.strip()

        if ' ' in key:
            _unparsed, key = key.rsplit(' ', 1)
            unparsed.append(_unparsed)

        for conversion in (int, float):
            try:
                value = conversion(value)
            except ValueError:
                pass
            else:
                break

        parsed[key] = value

    if unparsed:
        parsed['zb.unparsed'] = json.dumps(unparsed)
        parsed['zb.unparsed.line'] = buf

    return parsed
ENVELOPE_PARSERS.append(parse_comma_equals)


# NOTE: this is the "if all else fails" parser; should be appended last!
def parse_log_envelope(buf):
    params = {
        'action': 'log',
        'line': buf,
    }

    return params
ENVELOPE_PARSERS.append(parse_log_envelope)


def get_envelope(action, params, client_id=None, as_json=True):
    data = {
        'message_id': str(uuid.uuid4()),
        'timestamp': timestamp(),
        'action': action,
        'params': params,
    }

    if client_id:
        data.update({
            'client_id': client_id,
        })

    if as_json:
        return json.dumps(data)
    else:
        return data


def parse_buf(buf):
    """
    parse the given buffer into an envelope
    :param buf: string, may be in a parseable format
    :return: envelope
    """
    for parser in ENVELOPE_PARSERS:
        params = parser(buf)
        if params:
            return get_parsed_envelope(params)



"""
Connect to the pubsub engine
"""
import datetime
import logging
import math

from _ssl import SSLWantReadError

from zymbit import settings
from zymbit.exceptions import NotConnected, Disconnect
from zymbit.upstream import registration
from zymbit.upstream.ws import get_websocket
from zymbit.util.client import get_auth_token
from zymbit.util.envelope import get_envelope
from zymbit.util.statemachine import StateMachine, NO_SLEEP
from zymbit.util.time import now

NO_DELTA = datetime.timedelta(seconds=0)


class PubSubStateMachine(StateMachine):
    """
    State machine to keep connect to the pubsub engine

    This state machine handles bootstrapping a system when it's not yet
    registered and once registered, establish a persistent connection to
    the pubsub engine
    """
    def __init__(self, raise_exceptions=True, message_handler=None, subscriptions=None):
        super(PubSubStateMachine, self).__init__(raise_exceptions=raise_exceptions)

        self.message_handler = message_handler

        self.registration_retries = 0
        self.next_registration_attempt = None
        self.registration_retry_max_sleep = 3600  # sleep up to an hour

        self.subscriptions = subscriptions or []

        self.websocket = None

        # set last_read to instantiation time so that ping pong is played after
        # the connection has been established
        self.last_read = now()
        self.last_ping = self.last_read

        # play ping pong after a minute of silence
        self.ping_interval = datetime.timedelta(seconds=settings.PUBSUB_PING_INTERVAL)

    @property
    def logger(self):
        return logging.getLogger('{}.{}'.format(__name__, self.__class__.__name__))

    def send(self, envelope):
        if self.websocket is None:
            raise NotConnected()

        self.websocket.send(envelope)

    ##
    # State machine methods
    ##

    def init(self):
        if self.message_handler:
            self.message_handler(get_envelope('proxy', dict(routing_key='proxy.init')))

    def check_last_read(self):
        _now = now()

        next_ping_check = self.last_read + self.ping_interval
        if (next_ping_check - _now) < NO_DELTA:
            # only send pings once per max_silence_time
            next_ping = self.last_ping + self.ping_interval
            if (next_ping - _now) < NO_DELTA:
                self.logger.debug('sending ping')
                self.websocket.send(get_envelope('ping', {}))
                self.last_ping = _now

        # check if a re-connect is in order
        disconnect_time = self.last_read + (self.ping_interval * 3)
        if (disconnect_time - _now) < NO_DELTA:
            raise Disconnect()

    def connect(self):
        """
        Connects to the pubsub engine
        """
        self.websocket = get_websocket()

        # set last_read here so that we are not immediately disconnected by check_last_read()
        self.last_read = now()

    def disconnect(self):
        """
        Disconnects from the pubsub engine
        """
        if self.message_handler:
            self.message_handler(get_envelope('connection', dict(routing_key='connection.disconnected')))

        if self.websocket is None:
            return

        ws, self.websocket = self.websocket, None

        ws.close()

    def handle_message(self, buf):
        if self.message_handler:
            self.message_handler(buf)
        else:
            self.logger.info(repr(buf))

    def has_auth_token(self):
        """
        Checks whether this device has an auth token
        """
        return get_auth_token() is not None

    def listen(self):
        """
        Listens for upstream messages and sends up local messages
        """
        try:
            buf = self.websocket.recv()
        except SSLWantReadError:  # seems to be raised when there is no data
            buf = None

        if buf:
            self.last_read = now()
            self.handle_message(buf)

            return NO_SLEEP

        self.check_last_read()

    def register(self):
        """
        Registers the system with zymbit services
        """
        # check to see if a registration attempt should be made
        if self.next_registration_attempt:
            _now = now()
            # when there is a positive delta between now and the next registration attempt
            # simply return
            if (self.next_registration_attempt - _now) > NO_DELTA:
                return False

            self.next_registration_attempt = None

        registration.register()

        self.registration_retries = 0

    def registration_error(self):
        self.logger.exception(self.last_exception)

        self.registration_retries += 1
        sleep_time = min(math.pow(2, self.registration_retries), self.registration_retry_max_sleep)
        self.next_registration_attempt = now() + datetime.timedelta(seconds=sleep_time)

        self.logger.error('Registration error; next retry at {}'.format(self.next_registration_attempt))

    def subscribe(self):
        """
        Subscribes to desired streams
        """
        for subscription in self.subscriptions:
            if isinstance(subscription, dict):
                params = subscription
            else:
                params = dict(routing_key=subscription)

            envelope = get_envelope('subscribe', params=params)
            self.websocket.send(envelope)

    transitions = {
        StateMachine.start: {
            True: init,
        },
        init: {
            None: has_auth_token,
        },
        has_auth_token: {
            False: register,
            True: connect,
        },
        register: {
            None: connect,
            Exception: registration_error,
        },
        registration_error: {
            None: StateMachine.start,
        },
        connect: {
            None: subscribe,
            Exception: disconnect,
        },
        disconnect: {
            None: StateMachine.start,
            Exception: StateMachine.start,
        },
        subscribe: {
            None: listen,
            Exception: disconnect,
        },
        listen: {
            Exception: disconnect,
        },
    }



from __future__ import absolute_import

import inspect
import logging
import time

from .time import now, get_sleep_time

NO_SLEEP = '-- NO SLEEP --'


class StateMachine(object):
    transitions = {}

    def __init__(self, raise_exceptions=True):
        self._run = True
        self._state = self.start

        self.raise_exceptions = raise_exceptions
        self.loop_sleep_time = 1.0

        self.last_exception = None

        self._setup_transitions()
        self.logger.debug('transitions={}'.format(self.transitions))

    def _setup_transitions(self):
        # convert the transition functions into bound methods
        _transitions = {}
        for k, v in list(self.transitions.items()):
            bound_method = getattr(self, k.__name__)
            t_transitions = dict([(kk, getattr(self, vv.__name__)) for kk, vv in list(v.items())])

            _transitions[bound_method] = t_transitions

        self.transitions = _transitions

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def loop(self):
        result = None

        try:
            result = self._state()
        except Exception as exc:  # global exception catcher here to use for state transitions
            self.last_exception = exc

            result = exc
            if not inspect.isclass(exc):
                result = exc.__class__

            if self.raise_exceptions:
                raise
            else:
                self.logger.exception(exc)
        else:
            self.last_exception = None
        finally:
            transitions = self.transitions.get(self._state, {})

            for _result, _state in list(transitions.items()):
                if _result == result or inspect.isclass(_result) and inspect.isclass(result) and issubclass(result, _result):
                    self._state = _state

        return result

    def quit(self):
        self._run = False

    def run(self):
        while self._run:
            start = now()

            current_state = self._state
            result = self.loop()

            # only sleep when there is no state transition
            if current_state == self._state and result != NO_SLEEP:
                sleep_time = get_sleep_time(self.loop_sleep_time, start)
                # self.logger.debug('loop_sleep_time={}, sleep_time={}'.format(self.loop_sleep_time, sleep_time))
                time.sleep(sleep_time)

    def start(self):
        return True



"""
Client class for local Connect server
"""
import json
import logging
import random
import select
import socket

from zymbit import settings
from zymbit.exceptions import NotConnected
from zymbit.util.statemachine import StateMachine
from zymbit.util.time import interval


class LocalClient(StateMachine):
    buffer_size = 4096
    subscriptions = []

    def __init__(self, raise_exceptions=False, loop_sleep_time=None, subscriptions=None):
        super(LocalClient, self).__init__(raise_exceptions=raise_exceptions)

        self.socket = None

        self.loop_sleep_time = loop_sleep_time or self.loop_sleep_time
        self.subscriptions = subscriptions or self.subscriptions

    @property
    def logger(self):
        return logging.getLogger('{}.{}'.format(__name__, self.__class__.__name__))

    def connect(self):
        address = (settings.CONSOLE_MESSENGER_HOST, settings.CONSOLE_MESSENGER_PORT)

        self.logger.info('address={}'.format(address))

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect(address)
        self.socket.setblocking(0)

        self.logger.debug('connected to {}'.format(address))

    def disconnect(self):
        if self.socket is not None:
            self.socket.close()

    def handle_buf(self, buf):
        buf_utf8 = buf.decode('utf8')

        try:
            envelope = json.loads(buf_utf8)
            if envelope.get('params', {}).get('routing_key') == 'connection.connected':
                self.subscribe()
        except ValueError:
            pass

        self.handle_message(buf_utf8)

    def handle_message(self, buf):
        self.logger.info(buf)

    def listen(self):
        r, _, _ = select.select([self.socket], [], [], 0.01)
        if self.socket in r:
            buf = self.socket.recv(self.buffer_size)

            self.handle_buf(buf)

        self.publish()

    def publish(self):
        pass

    def send(self, buf):
        if self.socket is None:
            raise NotConnected()

        if not buf.endswith('\n'):
            buf = '{}\n'.format(buf)

        self.socket.send(buf)

    def subscribe(self):
        for subscription in self.subscriptions:
            self.send('action=subscribe,routing_key={}'.format(subscription))

    transitions = {
        StateMachine.start: {
            True: connect,
        },
        connect: {
            None: listen,
            Exception: disconnect,
        },
        disconnect: {
            None: StateMachine.start,
            Exception: StateMachine.start,
        },
        listen: {
            socket.error: disconnect,
            Exception: disconnect,
        },
    }


class ExampleClient(LocalClient):
    subscriptions = [
        '#',
    ]

    @interval(30.0)
    def publish(self):
        value = int(5 * random.random())
        data = 'key=foo,value={}'.format(value)

        self.send(data)


if __name__ == '__main__':
    import sys

    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

    client = ExampleClient()
    client.run()



import os

from zymbit import settings
from zymbit.upstream.api import ZymbitApi
from zymbit.util import get_device_meta
from zymbit.util.client import get_auth_path, get_client_info


def register():
    """
    Makes the bootstrap request upstream
    """
    post_data = {}
    post_data.update(get_device_meta())
    post_data.update(get_client_info())

    api = ZymbitApi()

    settings_dict = {}
    for item in dir(settings):
        settings_dict[item] = getattr(settings, item)

    endpoint = settings.REGISTER_ENDPOINT.format(**settings_dict)
    response = api.post(endpoint, data=post_data)
    response.raise_for_status()

    write_auth(response.json())


def write_auth(data):
    auth_token = data['grains']['zymbit']['zymbot']['auth_token']
    write_auth_token(auth_token)


def write_auth_token(auth_token):
    auth_path = get_auth_path()
    auth_root = os.path.dirname(auth_path)
    if not os.path.exists(auth_root):
        os.makedirs(auth_root)

    with open(auth_path, 'wb') as fh:
        fh.write(auth_token)



import functools
import requests

from zymbit import settings
from zymbit.util.client import get_auth_token
from zymbit.util.version import get_version


class ZymbitApi(object):
    ConnectionError = requests.ConnectionError
    HTTPError = requests.HTTPError

    def __init__(self):
        self.response = None

        self.session = requests.session()

    def __getattribute__(self, item):
        if item in ('delete', 'get', 'patch', 'post', 'put'):
            request = super(ZymbitApi, self).__getattribute__('request')
            return functools.partial(request, item)

        return super(ZymbitApi, self).__getattribute__(item)

    def request(self, method, endpoint, **kwargs):
        headers = kwargs.pop('headers', {})
        headers['User-Agent'] = 'Zymbit Connect {}'.format(get_version())

        if 'Authorization' not in headers:
            auth_token = get_auth_token()
            if auth_token:
                headers['Authorization'] = 'Token {}'.format(auth_token)

        kwargs['headers'] = headers

        if 'verify' not in kwargs:
            kwargs['verify'] = settings.CHECK_HOSTNAME

        api_url = settings.API_URL
        url = '{}{}'.format(api_url, endpoint)

        method_fn = getattr(self.session, method)
        try:
            self.response = method_fn(url, **kwargs)
        except requests.ConnectionError as exc:
            raise requests.ConnectionError('Unable to connect to url={}'.format(url))
        else:
            self.response.raise_for_status()

        return self.response



import json
import logging
import time

from zymbit.connect.pubsub import PubSubStateMachine, NotConnected
from zymbit.connect.server import ConsoleMessengerServer
from zymbit.util.envelope import parse_buf
from zymbit.util.statemachine import NO_SLEEP
from zymbit.util.time import get_sleep_time, now


class Proxy(object):
    def __init__(self):
        self.pubsub = PubSubStateMachine(raise_exceptions=False, message_handler=self.handle_pubsub_message)
        self.messenger_server = ConsoleMessengerServer(self.handle_console_message)

        # when set, this message sent to all messenger server clients
        self.initial_message = None

        self._run = True

    @property
    def logger(self):
        return logging.getLogger('{}.{}'.format(__name__, self.__class__.__name__))

    def handle_console_message(self, client, buf):
        for item in buf.split('\n'):
            self.handle_buf(client, item)

    def handle_buf(self, client, buf):
        try:
            envelope = parse_buf(buf)
        except:
            self.logger.warning('unable to parse buf={!r}'.format(buf))
            return

        # check to see if this is a connection notification

        if self.handle_console_connection(client, envelope) is True:
            return

        try:
            self.pubsub.send(envelope)
        except NotConnected as exc:
            self.logger.exception(exc)
            self.logger.error('unable to send pubsub buf={!r}, envelope={}'.format(buf, envelope))

    def handle_console_connection(self, client, envelope):
        data = json.loads(envelope)

        # nothing to do for disconnects
        if data['params'].get('routing_key') != 'connection.connected':
            return

        # nothing to do when there is no initial message
        if self.initial_message is None:
            return

        self.messenger_server.send(client, self.initial_message)

        return True

    def handle_pubsub_message(self, buf):
        if not buf.endswith('\n'):
            buf = '{}\n'.format(buf)

        for t_buf in buf.splitlines():
            data = json.loads(t_buf)

            if data.get('params', {}).get('routing_key') == 'connection.connected':
                self.initial_message = t_buf
            elif data.get('params', {}).get('routing_key') == 'connection.disconnected':
                self.initial_message = None

        try:
            self.messenger_server.broadcast(buf)
        except Exception as exc:
            self.logger.exception(exc)
            self.logger.error('unable to send messenger_server buf={!r}'.format(buf))

    def run(self):
        while self._run:
            start = now()

            pubsub_result = self.pubsub.loop()
            messenger_result = self.messenger_server.loop(select_timeout=0.01)

            if NO_SLEEP in (pubsub_result, messenger_result):
                continue

            time.sleep(get_sleep_time(1.0, start))
